from typing import List
from dataclasses import field

import pytest

from intents import Intent, Sys
from intents.model.intent import IntentParameterMetadata
from intents.service_connector import Prediction
from intents import language

def test_param_scheme_no_params():

    class no_param_intent(Intent):
        """Simple Intent with no parameters"""

    assert no_param_intent.parameter_schema() == {}

def test_param_scheme_with_params():

    class intent_with_params(Intent):
        """Intent with parameters"""
        required_param: Sys.Person
        required_list_param: List[Sys.Person]
        optional_param: Sys.Person = "John"
        optional_list_param: List[Sys.Person] = field(default_factory=lambda: ["Al", "John"])

    assert intent_with_params.parameter_schema() == {
        "required_param": IntentParameterMetadata(
            name="required_param",
            entity_cls=Sys.Person,
            is_list=False,
            required=True,
            default=None
        ),
        "required_list_param": IntentParameterMetadata(
            name="required_list_param",
            entity_cls=Sys.Person,
            is_list=True,
            required=True,
            default=None
        ),
        "optional_param": IntentParameterMetadata(
            name="optional_param",
            entity_cls=Sys.Person,
            is_list=False,
            required=False,
            default="John"
        ),
        "optional_list_param": IntentParameterMetadata(
            name="optional_list_param",
            entity_cls=Sys.Person,
            is_list=True,
            required=False,
            default=["Al", "John"]
        ),
    }

def test_param_scheme_invalid_list_default():
    
    with pytest.raises(ValueError):

        class intent_with_invalid_list_default(Intent):
            """Intent with parameters"""
            optional_list_param: List[Sys.Person] = 42

def test_fulfillment_messages():
    class MockPredictionImplementation(Prediction):
        @property
        def entity_mappings(self):
            return None

    mock_default_messages = [
        language.TextIntentResponse(choices=["If you like I can recommend you an hotel. Or I can send you some holiday pictures"])
    ]
    mock_rich_messages = [
        language.TextIntentResponse(choices=["I also like travels, how can I help you?"]),
        language.QuickRepliesIntentResponse(replies=["Recommend an hotel", "Send holiday photo"])
    ]
    mock_prediction = MockPredictionImplementation(
        intent_name='fake_intent_name',
        confidence=0.5,
        contexts={},
        parameters_dict={},
        fulfillment_messages={
            language.IntentResponseGroup.DEFAULT: mock_default_messages,
            language.IntentResponseGroup.RICH: mock_rich_messages
        },
        fulfillment_text="Fake fulfillment text"
    )

    class fake_intent(Intent):
        pass

    predicted = fake_intent.from_prediction(mock_prediction)

    assert predicted.fulfillment_messages() == mock_rich_messages
    assert predicted.fulfillment_messages(language.IntentResponseGroup.DEFAULT) == mock_default_messages
    assert predicted.fulfillment_messages(language.IntentResponseGroup.RICH) == mock_rich_messages



"""
Here we implement :class:`DialogflowEsConnector`, an implementation of
:class:`Connector` that allows Agents to operate on Dialogflow.
"""
import os
import logging
import tempfile
from dataclasses import dataclass, field
from typing import Union, Iterable, Dict

import google.auth.credentials
from google.cloud.dialogflow_v2.types import TextInput, QueryInput, EventInput
from google.cloud.dialogflow_v2.services.sessions import SessionsClient
from google.cloud.dialogflow_v2.services.agents import AgentsClient
from google.cloud.dialogflow_v2.types import DetectIntentResponse, RestoreAgentRequest
from google.protobuf.json_format import MessageToDict

from intents import Agent, Intent
from intents.service_connector import Connector, Prediction
from intents.connectors.dialogflow_es.auth import resolve_credentials
from intents.connectors.dialogflow_es.util import dict_to_protobuf
from intents.connectors.dialogflow_es import entities as df_entities
from intents.connectors.dialogflow_es import export as df_export
from intents.connectors.dialogflow_es.response_format import intent_responses
from intents.connectors.commons import WebhookConfiguration

logger = logging.getLogger(__name__)

RICH_RESPONSE_PLATFORMS = ["telegram", "facebook", "slack", "line", "hangouts"]

# Dialogflow makes use of Protobuffer for its data structures, and protobuf may be
# tricky to deal with. For instance, `MessageToDict` will convert snake_case to
# lowerCamelCase, so API is documented snake_case, protobuf is snake_case,
# dialogflow results are camelCase, protobuf converted to dict is camelCase
# (unless you use a flag, in that case it could also be snake_case) 💀 This is one
# of the reasons this library exists.

@dataclass
class DialogflowPrediction(Prediction):
    """
    This is an implementation of :class:`Prediction` that comes from Dialogflow.

    `DialogflowPredictions` are produced internally by
    :class:`DialogflowEsConnector`, and automatically used to instantiate
    Intents in :class:`DialogflowEsConnector.predict` and
    :class:`DialogflowEsConnector.trigger`. Probably you won't need to manually
    operate with Predictions.
    """
    df_response: DetectIntentResponse = None

    entity_mappings = df_entities.MAPPINGS

class DialogflowEsConnector(Connector):
    """
    This is an implementation of :class:`Connector` that enables Agents to
    work as Dialogflow projects.

    An Agent can be connected to Dialogflow by providing its :class:`Agent`
    class and service account credentials for the the Google Cloud project
    that hosts the Dialogflow ES agent:

    .. code-block:: python

        from example_agent import ExampleAgent
        from intents.connectors import DialogflowEsConnector
        df = DialogflowEsConnector('/path/to/your/service-account-credentials.json', ExampleAgent)

    The Connector can now be used, mainly to

    * Export the Agent with :meth:`DialogflowEsConnector.export`
    * Predict an utterance with :meth:`DialogflowEsConnector.predict`
    * Trigger an Intent with :meth:`DialogflowEsConnector.trigger`

    :param google_credentials: Path to service account JSON credentials, or a Credentials object
    :param agent_cls: The Agent to connect
    :param default_session: An arbitrary string to identify the conversation during predictions. Will be generated randomly if None
    :param dedefault_language: Default language to use during predictions
    :param rich_platforms: Platforms to include when exporting Rich response messages
    :param webhook_configuration: Webhook connection parameters
    """
    entity_mappings = df_entities.MAPPINGS
    rich_platforms: Iterable[str]
    webhook_configuration: WebhookConfiguration

    _credentials: google.auth.credentials.Credentials
    _session_client: SessionsClient

    def __init__(
        self,
        google_credentials: Union[str, google.auth.credentials.Credentials],
        agent_cls: type(Agent),
        default_session: str=None,
        default_language: str="en",
        rich_platforms: Iterable[str]=("telegram",),
        webhook_configuration: WebhookConfiguration=None
    ):
        super().__init__(agent_cls, default_session=default_session,
                         default_language=default_language)
        self._credentials = resolve_credentials(google_credentials)
        assert all([p in RICH_RESPONSE_PLATFORMS for p in rich_platforms])
        self._session_client = SessionsClient(credentials=self._credentials)
        self.rich_platforms = rich_platforms
        self.webhook_configuration = webhook_configuration

    @property
    def gcp_project_id(self) -> str:
        """
        Return the Google Cloud Project ID that is associated with the current Connection
        """
        return self._credentials.project_id

    def export(self, destination: str):
        agent_name = 'py-' + self.agent_cls.__name__
        return df_export.export(self, destination, agent_name)

    def upload(self):
        agents_client = AgentsClient(credentials=self._credentials)
        with tempfile.TemporaryDirectory() as tmp_dir:
            export_path = os.path.join(tmp_dir, 'agent.zip')
            self.export(export_path)
            with open(export_path, 'rb') as f:
                agent_content = f.read()
            restore_request = RestoreAgentRequest(
                parent=f"projects/{self.gcp_project_id}",
                agent_content=agent_content
            )
            agents_client.restore_agent(request=restore_request)


    def predict(self, message: str, session: str = None, language: str = None) -> Intent:
        if not session:
            session = self.default_session
        if not language:
            language = self.default_language

        text_input = TextInput(text=message, language_code=language)
        query_input = QueryInput(text=text_input)
        session_path = self._session_client.session_path(
            self.gcp_project_id, session)
        df_result = self._session_client.detect_intent(
            session=session_path,
            query_input=query_input
        )
        df_response = df_result

        prediction = _df_response_to_prediction(df_response)
        return self._prediction_to_intent(prediction)

    def trigger(self, intent: Intent, session: str=None, language: str=None) -> Intent:
        if not session:
            session = self.default_session
        if not language:
            language = self.default_language

        intent_name = intent.name
        event_name = Agent._event_name(intent_name)
        event_parameters = {}
        for param_name, param_metadata in intent.parameter_schema().items():
            param_mapping = df_entities.MAPPINGS[param_metadata.entity_cls]
            if param_name in intent.__dict__:
                param_value = intent.__dict__[param_name]
                event_parameters[param_name] = param_mapping.to_service(param_value)

        logger.info("Triggering event '%s' in session '%s' with parameters: %s",
                    event_name, session, event_parameters)
        if not event_parameters:
            event_parameters = {}

        event_input = EventInput(
            name=event_name,
            parameters=dict_to_protobuf(event_parameters),
            language_code=language
        )
        query_input = QueryInput(event=event_input)
        session_path = self._session_client.session_path(
            self.gcp_project_id, session)
        result = self._session_client.detect_intent(
            session=session_path,
            query_input=query_input
        )
        df_response = result

        prediction = _df_response_to_prediction(df_response)
        return self._prediction_to_intent(prediction)

    def _prediction_to_intent(self, prediction: Prediction) -> Intent:
        """
        Turns a Prediction object into an Intent object
        """
        intent_class: Intent = self.agent_cls._intents_by_name.get(prediction.intent_name)
        if not intent_class:
            raise ValueError(f"Prediction returned intent '{prediction.intent_name}', but this was not found in Agent definition. Make sure to restore a latest Agent export from `services.dialogflow_es.export.export()`. If the problem persists, please file a bug on the Intents repository.")
        return intent_class.from_prediction(prediction)

#
# Response to prediction
#

def _df_response_to_prediction(df_response: DetectIntentResponse) -> DialogflowPrediction:
    return DialogflowPrediction(  # pylint: disable=abstract-class-instantiated
        intent_name=df_response.query_result.intent.display_name,
        parameters_dict=MessageToDict(
            df_response._pb.query_result.parameters
        ),  # TODO: check types
        # TODO: model
        contexts=[MessageToDict(c)
                  for c in df_response._pb.query_result.output_contexts],
        confidence=df_response.query_result.intent_detection_confidence,
        fulfillment_messages=intent_responses(df_response),
        fulfillment_text=df_response.query_result.fulfillment_text,
        df_response=df_response
    )



"""
Here we define schemas and utilities to convert Dialogflow Response elements to
*Intents* native classes.

TODO: consider using same schemas (e.g. for response messages) in Agent export
"""
from typing import Dict, List
from collections import defaultdict
from abc import ABC, abstractmethod

from google.cloud.dialogflow_v2 import types as df_types
from google.protobuf.json_format import MessageToDict

from intents.language import IntentResponseGroup, IntentResponse, TextIntentResponse, QuickRepliesIntentResponse, CardIntentResponse

class DialogflowIntentResponse(ABC):
    """
    An interface to add IntentResponse de-serialization from messages in
    Dialogflow Responses. 
    """

    @classmethod
    def from_df_message(cls, df_message_data: df_types.intent.Intent.Message) -> IntentResponse:
        """
        Build the IntentResponse from a DialogflowResult
        """
        raise NotImplementedError()

class DialogflowTextResponse(TextIntentResponse, DialogflowIntentResponse):

    @classmethod
    def from_df_message(cls, df_message_data: df_types.intent.Intent.Message) -> TextIntentResponse:
        """
        .. code-block:: json

            {
                "text": {
                    "text": [
                        "Hello, Human"
                    ]
                }
            }
        """
        message_dict = MessageToDict(df_message_data._pb)
        assert isinstance(message_dict["text"], list)
        return TextIntentResponse(
            choices=message_dict["text"]
        )

class DialogflowQuickRepliesResponse(QuickRepliesIntentResponse, DialogflowIntentResponse):

    @classmethod
    def from_df_message(cls, df_message_data: df_types.intent.Intent.Message) -> TextIntentResponse:
        """
        .. code-block: json
            
            {
                "title": "Quick Replies",
                "quickReplies": [
                    "Recommend an hotel",
                    "Send holiday photo"
                ]
            }
        """
        message_dict = MessageToDict(df_message_data._pb)
        return QuickRepliesIntentResponse(
            replies=message_dict["quickReplies"]
        )

class DialogflowCardResponse(CardIntentResponse, DialogflowIntentResponse):

    @classmethod
    def from_df_message(cls, df_message_data: df_types.intent.Intent.Message) -> CardIntentResponse:
        message_dict = MessageToDict(df_message_data._pb)
        message_buttons = message_dict.get("buttons", [])
        card_link = message_buttons[0].get('postback') if message_buttons else None
        return CardIntentResponse(
            title=message_dict.get("title"),
            subtitle=message_dict.get("subtitle"),
            image=message_dict.get("imageUri"),
            link=card_link
        )

def build_response_message(df_message):
    """
    Build a fulfillment message from a protobuf structure, as it is found in a
    protobuf response (`query_result.fulfillment_messages`)
    """
    if df_message.text:
        return DialogflowTextResponse.from_df_message(df_message.text)
    if df_message.quick_replies:
        return DialogflowQuickRepliesResponse.from_df_message(df_message.quick_replies)
    if df_message.card:
        return DialogflowCardResponse.from_df_message(df_message.card)

    raise ValueError(f"Unsupported Fulfillment Message: {df_message}")

def intent_responses(df_response: df_types.DetectIntentResponse) -> Dict[IntentResponseGroup, List[IntentResponse]]:
    """
    Return the responses in the prediction as standard instances of
    :class:`IntentResponse` and grouped by response group
    (:var:`IntentResponseGroup.DEFAULT` or :var:`IntentResponseGroup.RICH`) 
    """
    result = defaultdict(list)
    for message in df_response.query_result.fulfillment_messages:
        if message.platform == df_types.intent.Intent.Message.Platform.PLATFORM_UNSPECIFIED:
            group = IntentResponseGroup.DEFAULT
        else:
            group = IntentResponseGroup.RICH

        result[group].append(build_response_message(message))
    return dict(result)



from unittest.mock import patch

from google.cloud.dialogflow_v2.types import DetectIntentResponse

from intents.connectors.dialogflow_es.connector import DialogflowEsConnector
from intents import language
from example_agent import ExampleAgent, travels

# A full Dialogflow Response, with
# - A text message in the DEFAULT platform
# - A text message in the TELEGRAM platform
# - A quick replies message in the TELEGRAM platform
df_response_quick_replies_serialized = b'\n-1cedb9e6-f958-437f-9299-74f966fbec62-9779ea79\x12\xa4\x03\n\x10i want to travel"\x00(\x012QIf you like I can recommend you an hotel. Or I can send you some holiday pictures:U\nS\nQIf you like I can recommend you an hotel. Or I can send you some holiday pictures:.\n*\n(I also like travels, how can I help you?0\x03:;\x1a7\n\rQuick Replies\x12\x12Recommend an hotel\x12\x12Send holiday photo0\x03Zl\nOprojects/learning-dialogflow/agent/intents/e3a1e749-be67-11eb-8ad8-bbef97dc13e7\x12\x19travels.user_wants_travele\x00\x00\x80?z\x02en'
df_response_quick_replies = DetectIntentResponse.deserialize(df_response_quick_replies_serialized)

# TODO: one with parameters and contexts

@patch("intents.connectors.dialogflow_es.connector.resolve_credentials")
@patch("intents.connectors.dialogflow_es.connector.EventInput")
@patch("intents.connectors.dialogflow_es.connector.QueryInput")
@patch("intents.connectors.dialogflow_es.connector.TextInput")
@patch("intents.connectors.dialogflow_es.connector.SessionsClient")
def test_predict(mock_df_client_class, *args):
    # TODO: this relies on the consistency between mock prediction and
    # ExampleAgent
    # TODO: this is more functional test than unit test...
    def mock_df_client():
        pass
    def mock_detect_intent(session, query_input):
        return df_response_quick_replies
    def mock_session_path(gcp_project_id, session):
        return 'fake-full-session-path'
    mock_df_client.detect_intent = mock_detect_intent
    mock_df_client.session_path = mock_session_path
    mock_df_client_class.return_value = mock_df_client

    df = DialogflowEsConnector('/fake/path/to/credentials.json', ExampleAgent)
    predicted = df.predict("A fake sentence")
    assert isinstance(predicted, travels.user_wants_travel)
    assert predicted.fulfillment_text == df_response_quick_replies.query_result.fulfillment_text
    assert predicted.prediction.fulfillment_messages == {
        language.IntentResponseGroup.DEFAULT: [
            language.TextIntentResponse(choices=["If you like I can recommend you an hotel. Or I can send you some holiday pictures"])
        ],
        language.IntentResponseGroup.RICH: [
            language.TextIntentResponse(choices=["I also like travels, how can I help you?"]),
            language.QuickRepliesIntentResponse(replies=["Recommend an hotel", "Send holiday photo"])
        ]
    }



"""
Some configuration classes are not service-specific. Currently the only one is
:class:`WebhookConfiguration`.
"""
from typing import Dict
from dataclasses import dataclass, field

@dataclass
class WebhookConfiguration:
    """
    Specifies connection parameters for the Agent's webhook. This is the
    endpoint that is called by services to fulfill an intent.
    """
    url: str
    headers: Dict[str, str] = field(default_factory=dict)



"""
Connectors allow *Intents* Agent definitions to operate with real cloud services
such as Dialogflow, Lex or Azure Bot Services. Currently, only one connector is
provided with this library, and this is for Dialogflow ES:
:mod:`intents.connectors.dialogflow_es.connector`.

.. note::

    Details about the Connector interface are only useful if you need to develop your own Service Connector (please consider raising a pull request if this is the case). If you just need to use the included Dialogflow Connector you can jump to its documentation page right away: :mod:`intents.connectors.dialogflow_es.connector`

Connectors are used to operate with the cloud version of the Agent, and
specifically to:

* Export an :class:`intents.Agent` in a format that is natively readable by the
  Service
* Predict User messages on the Cloud Agent
* Trigger Intents on the Cloud Agent

More details can be found in the :class:`Connector` interface.
"""
from uuid import uuid1
from abc import ABC, abstractmethod
from typing import Dict, Any, List
from dataclasses import dataclass

from intents import Intent, Agent
from intents.model.intent import IntentParameterMetadata
from intents.model.entity import SystemEntityMixin, _EntityMetaclass
from intents.language import IntentResponse, IntentResponseGroup

class EntityMapping(ABC):
    """
    An Entity Mapping is a (de-)serializer for predicted Entities.

    Most of the times a Mapping is not needed as the Entity can be mapped
    directly to its type (e.g. "3" -> `Number(3)`). However, prediction services
    such as Dialogflow may define system entities of structured types; a notable
    example is Dialogflow's *sys.person* entity, which is returned as `{"name":
    "John"}` and therefore needs custom logic to be mapped to `Person("John")`.
    This is modelled in :class:`intents.connectors.dialogflow_es.entities.PersonEntityMapping`.

    Another notable scenario is Date/Time objects. A mapping can be used to
    convert time strings from the Service format to python objects. For
    Dialogflow ES, this is modelled in  :class:`intents.connectors.dialogflow_es.entities.DateTimeEntityMapping`.
    """

    @property
    @abstractmethod
    def entity_cls(self) -> _EntityMetaclass:
        """
        This is the internal entity type that is being mapped.

        >>> mapping = StringEntityMapping(Sys.Integer, 'sys.number-integer')
        >>> mapping.entity_cls
        Sys.Integer
        """

    @property
    @abstractmethod
    def service_name(self) -> str:
        """
        This is the name of the Entity in the Prediction Service domain.

        >>> mapping = StringEntityMapping(Sys.Integer, 'sys.number-integer')
        >>> mapping.service_name
        'sys.number-integer'
        """

    @abstractmethod
    def from_service(self, service_data: Any) -> SystemEntityMixin:
        """
        De-serialize the Service representation of an Entity (typically the value
        that is returned at prediction time) to an instance of one of the internal Entity
        classes in :class:`intents.model.entities`

        :param service_data: A parameter value, as it is returned by the Service
                             in a prediction/trigger response
        :return: the parameter value, modelled as one of the System Entity classes
        """

    @abstractmethod
    def to_service(self, entity: SystemEntityMixin) -> Any:
        """
        Serialize a System Entity instance into a Service representation (typically,
        to be sent as a parameter of a trigger request)

        :param entity: the System Entity to serialize
        :return: the serialized Entity that can be sent to Service (e.g. in a trigger request)
        """

class StringEntityMapping(EntityMapping):
    """
    This is a basic :class:`EntityMapping` that reads values as they are sent
    by the prediction service (e.g. `"3"` -> `Sys.Integer("3")`), and serializes
    by simple string conversion (e.g. `Sys.Integer(3)` -> "3").

    The System Entity to use must be defined when instantiating the mapping, for
    instance:

    >>> StringEntityMapping(Sys.Integer, "sys.number-integer")
    """

    entity_cls: _EntityMetaclass = None
    service_name: str = None

    def __init__(self, entity_cls: _EntityMetaclass, service_name: str):
        """
        :param service_name: name of the Entity in the Prediction Service
        """
        self.entity_cls = entity_cls
        self.service_name = service_name

    def from_service(self, service_data: Any) -> SystemEntityMixin:
        return self.entity_cls(service_data)

    def to_service(self, entity: SystemEntityMixin) -> Any:
        return str(entity)

class ServiceEntityMappings(dict):
    """
    Models a list of entity mappings. In addition to a standard list, two things
    are added:

    * Dict-like access to retrieve a mapping given its internal type (e.g.
      `mappings[Sys.Integer]`)
    * Consistency check: a mapping list must cover all the entities defined in
      the framework (TODO)
    * Shortcut to define StringEntityMappings like `(Sys.Integer, 'sys.number-integer')` (TODO)
    """

    @classmethod
    def from_list(cls, mapping_list: List[EntityMapping]) -> "ServiceEntityMappings":
        result = cls()
        for mapping in mapping_list:
            if mapping in result:
                raise ValueError(f"Mapping {mapping} already defined in list: {mapping_list}")
            result[mapping.entity_cls] = mapping
        return result

@dataclass
class Prediction(ABC):
    """
    This class is meant to abstract prediction results from a generic Prediction
    Service. It uses names from Dialogflow, as it is currently the only
    supported service
    """
    intent_name: str
    confidence: str
    contexts: dict
    parameters_dict: dict
    fulfillment_messages: Dict[IntentResponseGroup, List[IntentResponse]]
    fulfillment_text: str = None

    @property
    @abstractmethod
    def entity_mappings(self) -> ServiceEntityMappings:
        """
        A Prediction subclass must know the Entity Mappings of its Prediction
        Service. :meth:`parameters` will use such mappings to convert parameters
        in `parameters_dict` to their generic `Sys.*` type.
        """

    def parameters(self, schema: Dict[str, IntentParameterMetadata]) -> Dict[str, Any]:
        """
        Cast parameters in `Prediction.parameters_dict` according to the given
        schema. This is necessary, because parameter types are not known at
        prediction time, and therefore they need to be determined by the
        corresponding :class:`Intent` class.
        """
        result = {}
        for param_name, param_value in self.parameters_dict.items():
            if param_name not in schema:
                raise ValueError(f"Found parameter {param_name} in Service Prediction, but Intent class does not define it.")
            param_metadata = schema[param_name]
            mapping = self.entity_mappings[param_metadata.entity_cls]
            try:
                if param_metadata.is_list:
                    if not isinstance(param_value, list):
                        raise ValueError(f"Parameter {param_name} is defined as List, but returned value is not of 'list' type: {param_value}")
                    result[param_name] = [mapping.from_service(x) for x in param_value]
                else:
                    result[param_name] = mapping.from_service(param_value)
            except Exception as exc:
                raise RuntimeError(f"Failed to match parameter '{param_name}' with value '{param_value}' against schema {schema}. See source exception above for details.") from exc

        return result

class Connector(ABC):

    agent_cls: type(Agent)
    default_session: str
    default_language: str

    def __init__(self, agent_cls: type(Agent), default_session: str=None, default_language: str="en"):
        """
        Connect the given Agent to a Prediction Service.

        :param agent_cls: The Agent to connect
        :param default_session: A default session ID (conversation channel) for predictions
        :param default_language: A default language for predictions
        """
        if not default_session:
            default_session = f"py-{str(uuid1())}"
        
        self.agent_cls = agent_cls
        self.default_session = default_session
        self.default_language = default_language

    @abstractmethod
    def predict(self, message: str, session: str=None, language: str=None) -> Intent:
        """
        Predict the given User message in the given session using the given
        language. When `session` or `language` are None, `predict` will use the
        default values that are specified in :meth:`__init__`.

        *predict* will return an instance of `Intent`, representing the intent
        as it was predicted by the service.

        >>> from intents.connectors import DialogflowEsConnector
        >>> from example_agent import ExampleAgent
        >>> df = DialogflowEsConnector('/path/to/service-account.json', ExampleAgent)
        >>> df_result = df.predict("Hi, my name is Guido")
        user_name_give(user_name='Guido')
        >>> df_result.user_name
        "Guido"
        >>> df_result.fulfillment_text
        "Hi Guido, I'm Bot"
        >>> df_result.confidence
        0.86

        :param message: The User message to predict
        :param session: Any string identifying a conversation
        :param language: A ISO 639-1 language code (e.g. "en")
        :return: An instance of the predicted Intent class
        """

    @abstractmethod
    def trigger(self, intent: Intent, session: str=None, language: str=None) -> Intent:
        """
        Trigger the given Intent in the given session using the given language.
        When `session` or `language` are None, `predict` will use the default
        values that are specified in :meth:`__init__`.

        >>> from intents.connectors import DialogflowEsConnector
        >>> from example_agent import ExampleAgent, smalltalk
        >>> df = DialogflowEsConnector('/path/to/service-account.json', ExampleAgent)
        >>> df_result = df.trigger(smalltalk.agent_name_give(agent_name='Alice'))
        agent_name_give(agent_name='Alice')
        >>> df_result.fulfillment_text
        "Howdy Human, I'm Alice"
        >>> df_result.confidence
        1.0

        :param intent: The Intent instance to trigger
        :param session: Any string identifying a conversation
        :param language: A ISO 639-1 language code (e.g. "en")
        :return: An instance of the triggered Intent class
        """

    @abstractmethod
    def upload(self):
        """
        Upload the connected Agent to the Prediction Service.
        """

    @abstractmethod
    def export(self, destination: str):
        """
        Export the connected Agent in a format that can be read and imported
        natively by the Prediction service. For instance, the Dialogflow service
        will produce a ZIP export that can be imported from the Dialogflow
        console.

        Note that you can also directly upload the Agent with :meth:`upload`.

        :param destination: destination path of the exported Agent
        """



"""
**Intent** language files have the following structure:

.. code-block:: yaml

    examples:
      - an example utterance
      - another example utterance
      - an example utterance with $foo{42} as a numeric parameter

    slot_filling_prompts:
      foo:
        - Tell me the value for "foo"

    responses:
      default:
        - text:
          - A plain text response
          - An alternative response
          - Another alternative, referencing $foo as a paramneter
      rich:
        - text:
          - A text response for rich clients
        - quick_replies:
          - a reply chip
          - another reply chip

Let's look at the sections of this file.

* **examples** contain example utterances that will be used to predict the given
  intent. If your intent has a parameter, it can be referenced as
  $parameter_name{example value}. You can omit this section if your intent is
  not meant to be predicted (some intents are trigger-only)
* **slot_filling_prompt** are used when your intent has a mandatory parameter,
  and this parameter could not be matched in the user message. These prompts
  will be used to ask the User about that parameter. You can omit this section
  if your intent has no mandatory parameters, or if you don't want to define
  custom prompts.
* **responses** contain messages that Agent will send to User in response to the
  Intent. Two response groups are available:

  * **default** can only contain plain-text messages. It is good practice to
    always provide text-only response for situations where rich ones can't be rendered,
    such as vocal assistants, smartphone notifications and such. The `text`
    response type is specified in :class:`TextIntentResponse`
  * **rich** responses allow some extra types:
    :class:`QuickRepliesIntentResponse`, :class:`ImageIntentResponse`,
    :class:`CardIntentResponse` and :class:`CustomPayloadIntentResponse`

"""

import os
import re
from enum import Enum
from dataclasses import dataclass
from typing import List, Dict, Union

import yaml

from intents.language import agent_language, LanguageCode
from intents.model.entity import _EntityMetaclass

#
# Example Utterances
#

class UtteranceChunk:
    """
    An Example Utterance can be seen as a sequence of Chunks, where each Chunk
    is either a mapped Entity, or a plain text string.
    """

@dataclass
class TextUtteranceChunk(UtteranceChunk):
    """
    An Utterance Chunk that is a static, plain text string.
    """
    text: str

@dataclass
class EntityUtteranceChunk(UtteranceChunk):
    """
    An Utterance Chunk that is a matched entity
    """
    entity_cls: _EntityMetaclass
    parameter_name: str
    parameter_value: str

# TODO: check that parameter_value is one of the entries in custom entities
RE_EXAMPLE_PARAMETERS = re.compile(r"\$(?P<parameter_name>[\w]+)\{(?P<parameter_value>[^\}]+)\}")

class ExampleUtterance(str):
    """
    One of the example Utterances of a given Intent.
    """
    
    # TODO: check for escape characters - intent is possibly intent_cls
    def __init__(self, example: str, intent: "intents.Intent"):
        self._intent = intent
        self.chunks() # Will check parameters
    
    def __new__(cls, example: str, intent: "intents.Intent"):
        return super().__new__(cls, example)

    def chunks(self):
        """
        Return the Utterance as a sequence of :class:`UtteranceChunk`. Each
        chunk is either a plain text string, or a mapped Entity.

        >>> utterance = ExampleUtterance("My name is $user_name{Guido}!", intents.user_gives_name)
        >>> utterance.chunks()
        [
            TextUtteranceChunk(text="My name is "),
            EntityUtteranceChunk(entity_cls=Sys.Person, parameter_name="user_name", parameter_value="Guido"),
            TextUtteranceChunk(text="!")
        ]

        TODO: handle escaping
        """
        parameter_schema = self._intent.parameter_schema()
        result = []
        last_end = 0
        for m in RE_EXAMPLE_PARAMETERS.finditer(self):
            m_start, m_end = m.span()
            m_groups = m.groupdict()
            if m_start > 0:
                result.append(TextUtteranceChunk(text=self[last_end:m_start]))
            
            if (parameter_name := m_groups['parameter_name']) not in parameter_schema:
                raise ValueError(f"Example '{self}' references parameter ${parameter_name}, but intent {self._intent.name} does not define such parameter.")
 
            entity_cls = parameter_schema[parameter_name].entity_cls
            result.append(EntityUtteranceChunk(
                entity_cls=entity_cls,
                parameter_name=m_groups['parameter_name'],
                parameter_value=m_groups['parameter_value']
            ))

            last_end = m_end

        last_chunk = TextUtteranceChunk(text=self[last_end:])
        if last_chunk.text:
            result.append(last_chunk)

        return result

#
# Responses
#

class IntentResponseGroup(Enum):
    """
    Intent responses are divided in groups. The same intent can be answered with
    a set of plain-text responses (:const:`IntentResponseGroup.DEFAULT`), or with
    rich content (:const:`IntentResponseGroup.RICH`) that includes cards, images
    and quick replies.
    """
    DEFAULT = "default"
    RICH = "rich"

class IntentResponse:
    """
    One of the Response Utterances of a given Intent.
    """

    @classmethod
    def from_yaml(cls, data: dict):
        """
        Instantiate an IntentResponse from language data, as it's found in its
        YAML file. Typically, IntentResponse is a dataclass and `data` is a dict
        of fields; however specific subclasses may override with custom
        parameters.
        """
        return cls(**data)

@dataclass(frozen=True)
class TextIntentResponse(IntentResponse):
    """
    A plain text response. The actual response is picked randomly from a pool of
    choices.

    In the YAML definition a text response can either be a string, as in

    .. code-block:: yaml

        responses:
          default:
            - text: This is a response

    Or a list of choices (the output fulfillment message will be chosen
    randomly among the different options)

    .. code-block:: yaml

        responses:
          default:
            - text:
              - This is a response
              - This is an alternative response
    """

    choices: List[str]

    @classmethod
    def from_yaml(cls, data: Union[str, List[str]]):
        if isinstance(data, str):
            return cls([data])

        assert isinstance(data, list)
        return cls(data)

@dataclass(frozen=True)
class QuickRepliesIntentResponse(IntentResponse):
    """
    A set of Quick Replies that can be used to answer the Intent. Each reply
    must be shorter than 20 characters.

    In the YAML definition a quick replies response can either be a string, as in

    .. code-block:: yaml

        rich:
          - quick_replies: Order Pizza

    Or a list of replies, that will be rendered as separate chips

    .. code-block:: yaml

        rich:
          - quick_replies:
            - Order Pizza
            - Order Beer
    """

    replies: List[str]

    def __post_init__(self):
        for rep in self.replies:
            if len(rep) > 20:
                raise ValueError(f"Quick Replies must be shorter than 20 chars. Quick reply '{rep}' is {len(rep)} chars long.")

    @classmethod
    def from_yaml(cls, data: Union[str, List[str]]):
        if isinstance(data, str):
            return cls([data])
        
        assert isinstance(data, list)
        return cls(data)

@dataclass(frozen=True)
class ImageIntentResponse(IntentResponse):
    """
    A simple image, defined by its URL and an optional title

    In the YAML definition an image response can either be a string with the
    image URL, as in

    .. code-block:: yaml

        rich:
          - image: https://example.com/image.png

    Or an object with the image URL and a title, as in

    .. code-block:: yaml

        rich:
          - image:
              url: https://example.com/image.png
              title: An example image
    """
    url: str
    title: str = None

    @classmethod
    def from_yaml(cls, data: Union[str, List[str]]):
        if isinstance(data, str):
            return cls(url=data)
        
        assert isinstance(data, dict)
        return cls(**data)

@dataclass(frozen=True)
class CardIntentResponse(IntentResponse):
    """
    A simple content card that can be rendered on many platforms.

    In the YAML, this is defined as

    Or an object with the image URL and a title, as in

    .. code-block:: yaml

        rich:
          - card:
              title: The card title
              subtitle: An optional subtitle
              image: https://example.com/image.jpeg
              link: https://example.com/
    """
    title: str
    subtitle: str = None
    image: str = None
    link: str = None

@dataclass(frozen=True)
class CustomPayloadIntentResponse(IntentResponse):
    """
    Custom Payloads are objects with arbitrary fields, they are supported by
    Dialogflow in every response group, including "Default". Currently they can
    only be defined in the YAML as free form payloads; support for marshalling
    or generation from code is expected in future developments.
    """

    name: str
    payload: dict

    @classmethod
    def from_yaml(cls, data: Dict[str, dict]):
        """
        In the YAML definition a custom payload is defined as follows

        .. code-block:: yaml

            rich:
              - custom:
                  custom_location:
                    latitude: 45.484907
                    longitude: 9.203299
                    name: Piazza Duca D'Aosta, Milano

        NOTE: while not currently enforced, consistency is expected between
        payload names and their fields. Future versions of the library will
        marshal custom payloads against dataclass schemas.
        """
        if not isinstance(data, dict):
            raise ValueError(f"A custom payload is expected to be a dict in the form 'payload_name: {{\"foo\": \"bar\"}}. Found: {data}")
        if len(data) != 1:
            raise ValueError(f"A custom payload is expected to contain a single key representing the payload name, mapping to its value (e.g. 'location: {{\"latitude\": 42, ...}}'). Found {len(data)} keys: {data.keys()}")

        payload_name = list(data.keys())[0]
        payload_content = list(data.values())[0]

        if not isinstance(payload_content, dict):
            raise ValueError(f"Custom payloads are expected to be dictionaries. {payload_name} has value: {payload_content}")

        return CustomPayloadIntentResponse(payload_name, payload_content)

@dataclass
class IntentLanguageData:
    """
    Language data for an Intent consists of three resources:

    * Example Utterances
    * Slot Filling Prompts
    * Responses

    **Example Utterances** are the messages that Agent will be trained on to
    recognize the Intent.

    **Responses**, intuitively, are the Agent's response messages that will be sent
    to User once the Intent is recognized. They are divided in groups: a
    :const:`IntentResponseGroup.DEFAULT` group (mandatory) can only contain plain
    text responses. A :const:`IntentResponseGroup.RICH` group can provide intent
    responses that include cards, images and quick replies.

    **Slot Filling Promps** are used to solve parameters that couldn't be tagged in
    the original message. For instance a `order_pizza` intent may have a
    `pizza_type` parameter. When User asks "I'd like a pizza" we want to fill
    the slot by asking "What type of pizza?". `slot_filling_prompts` will map
    parameters to their prompts: `{"pizza_type": ["What type of pizza?"]}`
    """
    example_utterances: List[ExampleUtterance]
    slot_filling_prompts: Dict[str, List[str]]
    responses: Dict[IntentResponseGroup, List[IntentResponse]]

#
# Language Data Loader
#

def intent_language_data(
    agent_cls: "intents.model.agent._AgentMetaclass",
    intent_cls: "intents.model.intent._IntentMetaclass",
    language_code: LanguageCode=None
) -> Dict[LanguageCode, IntentLanguageData]:
    try:
        language_folder = agent_language.agent_language_folder(agent_cls)

        if not language_code:
            result = {}
            for language_code in agent_cls.languages:
                language_data = intent_language_data(agent_cls, intent_cls, language_code)
                result[language_code] = language_data[language_code]
            return result

        if isinstance(language_code, str):
            language_code = LanguageCode(language_code)

        language_file = os.path.join(language_folder, language_code.value, f"{intent_cls.name}.yaml")
        if not os.path.isfile(language_file):
            raise ValueError(f"Language file not found for intent '{intent_cls.name}'. Expected path: {language_file}. Language files are required even if the intent doesn't need language; in this case, use an empty file.")
        
        with open(language_file, 'r') as f:
            language_data = yaml.load(f.read(), Loader=yaml.FullLoader)

        if not language_data:
            return IntentLanguageData([], {}, [])

        examples_data = language_data.get('examples', [])
        responses_data = language_data.get('responses', [])

        examples = [ExampleUtterance(s, intent_cls) for s in examples_data]
        responses = _build_responses(responses_data)
        
        language_data = IntentLanguageData(
            example_utterances=examples,
            slot_filling_prompts=language_data.get('slot_filling_prompts', {}),
            responses=responses
        )

        return {language_code: language_data}
    except Exception as e:
        raise RuntimeError(f"Failed to load language data for intent {intent_cls.name} (see stacktrace above for root cause).") from e


def _build_responses(responses_data: dict):
    result = {}

    response_group: str
    responses: List[dict]
    for response_group, responses in responses_data.items():
        try:
            response_group = IntentResponseGroup(response_group)
        except ValueError as exc:
            raise NotImplementedError(f"Unsupported Response Group '{response_group}' in 'responses'. Currently, only 'default' and 'rich' are supported") from exc

        result[response_group] = []
        for r in responses:
            assert len(r) == 1
            for r_type, r_data in r.items():
                if response_group == IntentResponseGroup.DEFAULT and r_type != 'text':
                    raise ValueError(f"Message type {r_type} found in response group 'default'. Only 'text' type is allowed in 'default': please define the additional 'rich' response group to use rich responses.")

                if r_type == 'text':
                    result[response_group].append(TextIntentResponse.from_yaml(r_data))
                elif r_type == 'quick_replies':
                    result[response_group].append(QuickRepliesIntentResponse.from_yaml(r_data))
                elif r_type == 'image':
                    result[response_group].append(ImageIntentResponse.from_yaml(r_data))
                elif r_type == 'card':
                    result[response_group].append(CardIntentResponse.from_yaml(r_data))
                elif r_type == 'custom':
                    result[response_group].append(CustomPayloadIntentResponse.from_yaml(r_data))
                else:
                    raise NotImplementedError(f"Unsupported response type '{r_type}'. Currently, only 'text' is supported")
                
    return result



import os
import sys
import logging
from typing import List

from intents.language.language_codes import LanguageCode, LANGUAGE_CODES

logger = logging.getLogger(__name__)

def agent_language_folder(agent_cls: "agent._AgentMetaclass") -> str:
    main_agent_package_name = agent_cls.__module__.split('.')[0]
    main_agent_package = sys.modules[main_agent_package_name]
    if '__path__' not in main_agent_package.__dict__:
        # TODO: try workdir or something...
        logger.warning("Agent %s doesn't seem to be defined within a package. Language data will not be loaded.", agent_cls)
        return [], []

    agent_folder = main_agent_package.__path__[0]
    language_folder = os.path.join(agent_folder, 'language')
    if not os.path.isdir(language_folder):
        raise ValueError(f"No language folder found for agent {agent_cls} (expected: {language_folder})")

    return language_folder

def agent_supported_languages(agent_cls: "agent._AgentMetaclass") -> List[LanguageCode]:
    result = []
    
    language_folder = agent_language_folder(agent_cls)
    for f in os.scandir(language_folder):
        if f.is_dir() and not f.name.startswith('.')  and not f.name.startswith('_'):
            if f.name in LANGUAGE_CODES:
                result.append(LanguageCode(f.name))
            else:
                logger.warning("Unrecognized language code: '%s' (must be one of %s). Skipping language data.", f.name, LANGUAGE_CODES)
        
    return result



"""
Here we define classes that marshal the output format of Dialogflow Agents. We use
these to generate a set of JSON files representing the agent, in a format that is compatible
with the ones you get from Settings > Export in the Dialogflow console.

A Dialogflow Agent is represented by a compressed folder that contains the Agent
metadata, a set of Intents and a set of Entities.

Agent **metadata** are stored in a `agent.json` file. A complementary
`package.json` contains a standard `{"version": "1.0.0"}`. TODO: model metadata.

Intents can be found in the `intents/` folder. Each **intent** is defined by 2 files:

* `intents/<INTENT_NAME>.json` contains the Intent definition. This is modelled in :class:`Intent`
* `intents/<INTENT_NAME>_usersays.json` contains the list of example utterances. Each
  utterance is modelled in :class:`IntentUsersays`

Intents are modelled in the official client as Protobuf structures
(https://cloud.google.com/dialogflow/es/docs/reference/rpc/google.cloud.dialogflow.v2beta1#intent);
however, they seem to have little in common with the JSON output format.

Entities can be found in the `entities/` folder. Each **entity** is defined by 2
files:

* `entities/<ENTITY_NAME>.json` contains the Entity definition, this is modelled
  in :class:`Entity`
* `entities/<ENTITY_NAME>_entries.json` contains the entity values and their synonyms
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Union, Dict

#
# agent.json
#

@dataclass
class AgentGoogleAssistantOauthLinking:
    required: bool = False
    providerId: str = ""
    authorizationUrl: str = ""
    tokenUrl: str = ""
    scopes: str = ""
    privacyPolicyUrl: str = ""
    grantType: str = "AUTH_CODE_GRANT"

@dataclass
class AgentGoogleAssistant:
    project: str
    oAuthLinking: AgentGoogleAssistantOauthLinking
    googleAssistantCompatible: bool = False
    welcomeIntentSignInRequired: bool = False
    startIntents: List = field(default_factory=list)
    systemIntents: List = field(default_factory=list)
    endIntentIds: List = field(default_factory=list)
    voiceType: str = "MALE_1"
    capabilities: List = field(default_factory=list)
    env: str = ""
    protocolVersion: str = "V2"
    autoPreviewEnabled: bool = False
    isDeviceAgent: bool = False

@dataclass
class AgentWebhook:
    url: str = ""
    username: str = ""
    headers: Dict[str, str] = field(default_factory=dict)
    available: bool = False
    useForDomains: bool = False
    cloudFunctionsEnabled: bool = False
    cloudFunctionsInitialized: bool = False

@dataclass
class Agent:
    displayName: str
    webhook: AgentWebhook
    googleAssistant: AgentGoogleAssistant
    description: str = ""
    language: str = "en"
    shortDescription: str = ""
    examples: str = ""
    linkToDocs: str = ""
    disableInteractionLogs: bool = False
    disableStackdriverLogs: bool = True
    defaultTimezone: str = "Europe/Rome" # TODO: check
    isPrivate: bool = True
    mlMinConfidence: float = 0.3
    supportedLanguages: List[str] = field(default_factory=list)
    onePlatformApiVersion: str = "v2" # TODO: check
    secondaryKey: str = "" # TODO: check
    analyzeQueryTextSentiment: bool = False
    enabledKnowledgeBaseNames: List = field(default_factory=list)
    knowledgeServiceConfidenceAdjustment: float = -0.4
    dialogBuilderMode: bool = False
    baseActionPackagesUrl: str = ""

#
# entities/<ENTITY_NAME>.json
#

@dataclass
class Entity:
    id: str
    name: str
    isOverridable: bool = True
    isEnum: bool = False
    isRegexp: bool = False
    automatedExpansion: bool = False
    allowFuzzyExtraction: bool = False


#
# entities/<ENTITY_NAME>_entries_en.json
#

@dataclass
class EntityEntry:
    value: str
    synonyms: List[str]

#
# intents/<INTENT_NAME>.json
#

@dataclass
class AffectedContext:
    """
    This is the name of output contexts in the DF Intent definition
    """
    name: str
    lifespan: int

@dataclass
class Prompt:
    value: str
    lang: str

@dataclass
class Parameter:
    id: str
    name: str
    required: bool
    dataType: str
    value: str
    defaultValue: str = ""
    isList: bool = False
    prompts: List[Prompt] = field(default_factory=list)
    promptMessages: List = field(default_factory=list)
    noMatchPromptMessages: List = field(default_factory=list)
    noInputPromptMessages: List = field(default_factory=list)
    outputDialogContexts: List = field(default_factory=list)

class ResponseMessageTypes(Enum):
    TEXT = "0"
    CARD = "1"
    QUICK_REPLIES = "2"
    IMAGE = "3"
    CUSTOM = "4"

@dataclass
class ResponseMessage:
    lang: str
    type: str = "0" # TODO: use Enum with custom asdict() as string
    title: str = ""
    textToSpeech: str = ""
    condition: str = ""
    platform: str = None # TODO: model different platforms


@dataclass
class TextResponseMessage(ResponseMessage):
    speech: List[str] = ""
    type: str = "0"   # TODO: make constant, included in asdict()

@dataclass
class QuickRepliesResponseMessage(ResponseMessage):
    replies: List[str] = field(default_factory=list)
    title: str = "Quick Replies"
    type: str = "2"   # TODO: make constant, included in asdict()

@dataclass
class ImageResponseMessage(ResponseMessage):
    imageUrl: str = ""
    title: str = ""
    type: str = "3"

@dataclass
class CardResponseMessageButton:
    text: str
    postback: str = None

@dataclass
class CardResponseMessage(ResponseMessage):
    title: str = ""
    subtitle: str = ""
    imageUrl: str = ""
    buttons: List[CardResponseMessageButton] = None
    type: str = "1"

@dataclass
class CustomPayloadResponseMessage(ResponseMessage):
    payload: Dict[str, dict] = field(default_factory=dict)
    type: str = "4"


@dataclass
class Response:
    affectedContexts: List[AffectedContext]
    parameters: List[Parameter]
    messages: List[ResponseMessage]
    speech: List = field(default_factory=list)
    resetContexts: bool = False
    action: str = ""

@dataclass
class Event:
    name: str

@dataclass
class Intent:
    id: str
    name: str
    responses: List[Response]
    auto: bool = True
    contexts: List[str] = field(default_factory=list)
    priority: int = 500000
    webhookUsed: bool = True
    webhookForSlotFilling: bool = True
    fallbackIntent: bool = False
    events: List[Event] = field(default_factory=list)
    conditionalResponses: List = field(default_factory=list)
    condition: str = ""
    conditionalFollowupEvents: List = field(default_factory=list)

#
# intents/<INTENT_NAME>_usersays.json
#

class UsersaysChunk:
    pass

@dataclass
class UsersaysEntityChunk(UsersaysChunk):
    text: str
    meta: str
    alias: str
    userDefined: bool


@dataclass
class UsersaysTextChunk(UsersaysChunk):
    text: str
    userDefined: bool


@dataclass
class IntentUsersays:
    id: str
    lang: str
    data: List[Union[UsersaysTextChunk, UsersaysEntityChunk]]
    isTemplate: bool = False
    count: int = 0
    updated: int = 0



"""
An **intent** is a categorical representation of the User intention in a single
conversation turn. For instance, utterances like "I want a pizza", "I'd like to
order a pizza" and such, could be mapped to a single `order_pizza` intent.

Your agent will typically define a number of *intents*, representing all the
types of messages the Agent can understand and answer to. This is done by
defining :class:`Intent` sub-classes and their language resources (see
:mod:`intents.language`), and registering them to an :class:`intents.Agent`
class with :meth:`intents.Agent.register`.
"""

import re
import logging
import dataclasses
from dataclasses import dataclass, is_dataclass
from typing import List, Dict, Any, _GenericAlias

from intents.model import context, event, entity
from intents import language

logger = logging.getLogger(__name__)

#
# Intent
#

@dataclass
class IntentParameterMetadata:
    name: str
    entity_cls: entity._EntityMetaclass
    is_list: bool
    required: bool
    default: Any

class _IntentMetaclass(type):

    name: str = None
    input_contexts: List[context._ContextMetaclass] = None
    output_contexts: List[context.Context] = None
    events: List[event.Event] = None # TODO: at some point this may contain strings

    def __new__(cls, name, bases, dct):
        result_cls = super().__new__(cls, name, bases, dct)

        # Do not process Intent base class
        if name == 'Intent':
            assert not bases
            return result_cls

        if not result_cls.name:
            result_cls.name = _intent_name_from_class(result_cls)
        else:
            is_valid, reason = _is_valid_intent_name(result_cls.name)
            if not is_valid:
                raise ValueError(f"Invalid intent name '{result_cls.name}': {reason}")

        if not result_cls.input_contexts:
            result_cls.input_contexts = []
        if not result_cls.output_contexts:
            result_cls.output_contexts = []

        # TODO: check that custom parameters don't overlap Intent fields
        # TODO: check language data
        # language.intent_language_data(cls, result) # Checks that language data is existing and consistent

        events = [_system_event(result_cls.name)]
        for event_cls in result_cls.__dict__.get('events', []):
            events.append(event_cls)
        result_cls.events = events

        if not is_dataclass(result_cls):
            result_cls = dataclass(result_cls)

        # Check parameters
        result_cls.parameter_schema()

        return result_cls

class Intent(metaclass=_IntentMetaclass):
    """
    Represents a predicted intent. This is also used as a base class for the
    intent classes that model a Dialogflow Agent in Python code.

    In its simplest form, an Intent can be defined as follows:

    .. code-block:: python

        from intents import Intent

        class user_says_hello(Intent):
            \"\"\"A little docstring for my Intent\"\"\"

    *Intents* will then look for language resources in the folder where your
    Agent class is defined, and specifically in
    `language/<LANGUAGE-CODE>/user_says_hello.yaml`. More details in
    :mod:`intents.language`.

    Intents can be more complex than this, for instance:

    .. code-block:: python

        from dataclasses import dataclass
        from intents import Intent, Sys

        @dataclass
        class user_says_hello(Intent):
            \"\"\"A little docstring for my Intent\"\"\"

            user_name: Sys.Person

            name = "hello_custom_name"
            input_contexts = [a_context]
            input_contexts = [a_context(2), another_context(1)]

    This Intent has a custom name (i.e. will appear as "hello_custom_name" when
    exported to Dialogflow), will be predicted only when `a_context` is active,
    and will spawn `a_context`, lasting 2 conversation turns, and
    `another_context` lasting only 1 conversation turn.

    Most importantly, this intent has a `user_name` **parameter** of type
    :class:`Sys.Person` (check out :class:`intents.model.entity.Sys` for available system
    entities). With adequate examples in its language file, it will be able to
    match utterances like "Hello, my name is John", tagging "John" as an Entity.
    When a connector is instantiated, predictions will look like this:

    >>> predicted = connector.predict("My name is John")
    user_says_hello(user_name="John") predicted.user_name "John"
    predicted.fulfillment_text "Hi John, I'm Agent"

    Last, we notice the **@dataclass** decorator. This isn't really needed for
    the Intent to work, but adding it will have your IDE recognize the Intent
    class as a dataclass: you want autocomplete and static type checking when
    working with hundreds of intents in the same project.
    """
    # TODO: check parameter names: no underscore, no reserved names, max length

    name: str = None
    input_contexts: List[context._ContextMetaclass] = None
    output_contexts: List[context.Context] = None
    events: List[event.Event] = None # TODO: at some point this may contain strings

    # A :class:`Connector` provides this
    prediction: 'intents.Prediction'

    @property
    def confidence(self) -> float:
        return self.prediction.confidence

    @property
    def contexts(self) -> list:
        return self.prediction.contexts

    @property
    def fulfillment_text(self) -> str:
        return self.prediction.fulfillment_text

    def fulfillment_messages(
        self,
        response_group: "language.IntentResponseGroup"=language.IntentResponseGroup.RICH
    ) -> List["language.IntentResponse"]:
        """
        Return a list of fulfillment messages that are suitable for the given
        Response Group. The following scenarios may happen:

        * :class:`language.IntentResponseGroup.DEFAULT` is requested -> Message
          in the `DEFAULT` group will be returned
        * :class:`language.IntentResponseGroup.RICH` is requested

            * `RICH` messages are defined -> `RICH` messages are returned
            * No `RICH` message is defined -> `DEFAULT` messages are returned

        If present, messages in the "rich" group will be returned:

        >>> result.fulfillment_messages()
        [TextIntentResponse(choices=['I like travelling too! How can I help?']),
         QuickRepliesIntentResponse(replies=['Recommend a hotel', 'Send holiday photo', 'Where the station?'])]
         
        Alternatively, I can ask for plain-text default messages:

        >>> from intents.language import IntentResponseGroup
        >>> result.fulfillment_messages(IntentResponseGroup.DEFAULT)
        [TextIntentResponse(choices=['Nice, I can send you holiday pictures, or recommend an hotel'])]
        
        """
        if response_group == language.IntentResponseGroup.RICH and \
           not self.prediction.fulfillment_messages.get(response_group):
            response_group = language.IntentResponseGroup.DEFAULT

        return self.prediction.fulfillment_messages.get(response_group, [])

    @classmethod
    def parameter_schema(cls) -> Dict[str, IntentParameterMetadata]:
        """
        Return a dict representing the Intent parameter definition. A key is a
        parameter name, a value is a :class:`IntentParameterMetadata` object.

        TODO: consider computing this in metaclass to cache value and check types
        """
        result = {}
        for param_field in cls.__dict__['__dataclass_fields__'].values():
            # List[...]
            if isinstance(param_field.type, _GenericAlias):
                if param_field.type.__dict__.get('_name') != 'List':
                    raise ValueError(f"Invalid typing '{param_field.type}' for parameter '{param_field.name}'. Only 'List' is supported.")

                if len(param_field.type.__dict__.get('__args__')) != 1:
                    raise ValueError(f"Invalid List modifier '{param_field.type}' for parameter '{param_field.name}'. Must define exactly one inner type (e.g. 'List[Sys.Integer]')")
                
                # From here on, check the inner type (e.g. List[Sys.Integer] -> Sys.Integer)
                entity_cls = param_field.type.__dict__.get('__args__')[0]
                is_list = True
            else:
                entity_cls = param_field.type
                is_list = False

            required = True
            default = None
            if not isinstance(param_field.default, dataclasses._MISSING_TYPE):
                required = False
                default = param_field.default
            if not isinstance(param_field.default_factory, dataclasses._MISSING_TYPE):
                required = False
                default = param_field.default_factory()

            if not required and is_list and not isinstance(default, list):
                raise ValueError(f"List parameter has non-list default value in intent {cls}: {param_field}")

            result[param_field.name] = IntentParameterMetadata(
                name=param_field.name,
                entity_cls=entity_cls,
                is_list=is_list,
                required=required,
                default=default
            )

        return result

    @classmethod
    def from_prediction(cls, prediction: 'intents.Prediction') -> 'Intent':
        """
        Build an :class:`Intent` class from a :class:`Prediction`. In practice:

        #. Match parameters givent the Intent schema
        #. Instantiate the Intent
        #. Set the `prediction` field on the instantiated Intent.

        Note that this method is mostly for internal use, *connectors* will call
        it for you.
        """
        try:
            parameters = prediction.parameters(cls.parameter_schema())
        except ValueError as exc:
            raise ValueError(f"Failed to match parameters for Intent class '{cls}'. Prediction: {prediction}") from exc

        result = cls(**parameters)
        result.prediction = prediction
        return result

def _is_valid_intent_name(candidate_name):
    if re.search(r'[^a-zA-Z_\.]', candidate_name):
        return False, "must only contain letters, underscore or dot"

    if candidate_name.startswith('.') or candidate_name.startswith('_'):
        return False, "must start with a letter"

    if "__" in candidate_name:
        return False, "must not contain __"

    return True, None

def _intent_name_from_class(intent_cls: _IntentMetaclass) -> str:
    full_name = f"{intent_cls.__module__}.{intent_cls.__name__}"
    if "__" in full_name:
        logger.warning("Intent class '%s' contains repeated '_'. This is reserved: repeated underscores will be reduced to one, this may cause unexpected behavior.")
    full_name = re.sub(r"_+", "_", full_name)
    return ".".join(full_name.split(".")[-2:])

def _system_event(intent_name: str) -> str:
    """
    Generate the default event name that we associate with every intent.

    >>> _event_name('test.intent_name')
    'E_TEST_INTENT_NAME'
    """
    # TODO: This is only used in Dialogflow -> Deprecate and move to DialogflowConnector
    event_name = "E_" + intent_name.upper().replace('.', '_')
    return event.SystemEvent(event_name)



import pytest
from intents.language.intent_language import TextIntentResponse, QuickRepliesIntentResponse, ImageIntentResponse, CardIntentResponse, CustomPayloadIntentResponse

def test_text_intent_response_string():
    text_response_instance = TextIntentResponse(["ciao"])
    text_response_from_yaml = TextIntentResponse.from_yaml("ciao")
    assert text_response_instance == text_response_from_yaml
    assert text_response_from_yaml.choices == ["ciao"]

def test_text_intent_response_list():
    text_response_instance = TextIntentResponse(["pippo", "franco"])
    text_response_from_yaml = TextIntentResponse.from_yaml(["pippo", "franco"])
    assert text_response_instance == text_response_from_yaml
    assert text_response_from_yaml.choices == ["pippo", "franco"]

def test_quick_replies_response_string():
    quick_replies_instance = QuickRepliesIntentResponse(["ciao"])
    quick_replies_from_yaml = QuickRepliesIntentResponse.from_yaml("ciao")
    assert quick_replies_instance == quick_replies_from_yaml
    assert quick_replies_from_yaml.replies == ["ciao"]

def test_quick_replies_response_list():
    quick_replies_response_instance = QuickRepliesIntentResponse(["pippo", "franco"])
    quick_replies_response_from_yaml = QuickRepliesIntentResponse.from_yaml(["pippo", "franco"])
    assert quick_replies_response_instance == quick_replies_response_from_yaml
    assert quick_replies_response_from_yaml.replies == ["pippo", "franco"]

def test_quick_replies_response_too_long():
    with pytest.raises(ValueError):
        QuickRepliesIntentResponse(["pippo", "franco", "chi chi chi, co co co"])
        
    with pytest.raises(ValueError):
        QuickRepliesIntentResponse.from_yaml(["chi chi chi, co co co", "pippo", "franco"])

    with pytest.raises(ValueError):
        QuickRepliesIntentResponse.from_yaml("chi chi chi, co co co")

def test_image_response_string():
    image_response_instance = ImageIntentResponse(url="https://example.com/image.png")
    image_response_from_yaml = ImageIntentResponse.from_yaml("https://example.com/image.png")
    assert image_response_instance == image_response_from_yaml

def test_image_response_dict():
    image_response_instance = ImageIntentResponse(url="https://example.com/image.png", title="A title")
    image_response_from_yaml = ImageIntentResponse.from_yaml({"url": "https://example.com/image.png", "title": "A title"})
    assert image_response_instance == image_response_from_yaml

def test_custom_payload_response():
    custom_payload_instance = CustomPayloadIntentResponse("any_name", {"foo": "bar"})
    custom_payload_from_yaml = CustomPayloadIntentResponse.from_yaml({
        "any_name": {"foo": "bar"}
    })
    assert custom_payload_instance == custom_payload_from_yaml

def test_card_intent_response():
    card_reponse_instance = CardIntentResponse("Card Title", "a subtitle...")
    card_reponse_from_yaml = CardIntentResponse.from_yaml({
        "title": "Card Title",
        "subtitle":"a subtitle..."
    })
    assert card_reponse_instance == card_reponse_from_yaml

def test_custom_payload_not_dict_data():
    with pytest.raises(ValueError):
        CustomPayloadIntentResponse.from_yaml("foo")

def test_custom_payload_not_dict_value():
    with pytest.raises(ValueError):
        CustomPayloadIntentResponse.from_yaml({"foo": "bar"})

def test_custom_payload_multiple_keys():
    with pytest.raises(ValueError):
        CustomPayloadIntentResponse.from_yaml({
            "any_name": {"foo": "bar"},
            "another_name": {"not": "allowed"}
        })



"""
The :class:`Agent` base class is the entry point of your project. You will
subclass it when defining your own Agent, and will later :meth:`Agent.register`
Intent classes and other resources into it.

Once the Agent is defined, you will connect it to a cloud service with a
:class:`Connector`, to make prediction and trigger requests.
"""

import re
import logging
import inspect
from types import ModuleType
from typing import List, Dict, Union
from dataclasses import dataclass

from intents.model.intent import Intent, _IntentMetaclass
from intents.model.entity import EntityMixin, SystemEntityMixin, _EntityMetaclass
from intents.model.context import Context, _ContextMetaclass
from intents.model.event import _EventMetaclass
from intents import language

logger = logging.getLogger(__name__)

class _AgentMetaclass(type):

    languages: List[language.LanguageCode] = None

    def __new__(cls, name, bases, dct):
        result_cls = super().__new__(cls, name, bases, dct)
        
        if name == 'Agent':
            assert not bases
            return result_cls

        if not result_cls.languages:
            result_cls.languages = language.agent_supported_languages(result_cls)

        languages = []
        for lan in result_cls.languages:
            if isinstance(lan, language.LanguageCode):
                languages.append(lan)
            elif isinstance(lan, str):
                languages.append(language.LanguageCode(lan))
            else:
                raise ValueError(f"Unsupported language '{lan}' for Agent '{result_cls}'. Must be a value of 'intents.LanguageCode'")
        result_cls.languages = languages

        return result_cls

class Agent(metaclass=_AgentMetaclass):
    """
    As the name suggests, Agent is the base class that models an Agent
    definition within the *Intents* framework.

    Typically, you will define a single Agent class in your project, that could
    be as simple as this:

    .. code-block:: python

        from intents import Agent

        class MyAgent(Agent):
            \"\"\"A little docstring for your Agent\"\"\"

    You can optionally define the languages that you intend to support.
    *Intents* will look for language resources based on the `language` class
    variable:

    .. code-block:: python

        class MyAgent(Agent):
            \"\"\"A little docstring for your Agent\"\"\"
            languages = ["en", "it"]

    Languages are values from :class:`intents.language.LanguageCode`. If omitted,
    *Intents* will discover language resources by itself.

    You won't do much more with your *Agent* class, other than registering
    intents and resources with :meth:`Agent.register`, or passing it to a
    :class:`Connector` to make predictions.
    """

    languages: List[language.LanguageCode] = None

    intents: List[Intent] = None
    _intents_by_name: Dict[str, Intent] = None
    _intents_by_event: Dict[str, Intent] = None
    _entities_by_name: Dict[str, _EntityMetaclass] = None
    _contexts_by_name: Dict[str, _ContextMetaclass] = None
    _events_by_name: Dict[str, _EventMetaclass] = None

    @classmethod
    def register(cls, resource: Union[_IntentMetaclass, ModuleType]):
        """
        Register the given resource in Agent. The resource could be:

        * An :class:`Intent`
        * A module. In this case, the module is scanned (non recursively) for
          Intents, and each Intent is added individually

        This is how you register a **single intent**:

        .. code-block:: python

            from intents import Agent, Intent

            class MyAgent(Agent):
                pass

            @dataclass
            class my_test_intent(Intent):
                \"\"\"A little docstring for my Intent...\"\"\"
                a_parameter: str
                another_parameter: str

            MyAgent.register(my_test_intent)

        Alternatively, you can register a **whole module** containing Intents.
        This is how you register all the intents that are defined in the
        `smalltalk` module of `example_agent`:

        .. code-block:: python

            from example_agent import smalltalk

            class MyAgent(Agent):
                pass

            MyAgent.register(smalltalk)

        Note that together with each Intent, the Agent will register all of its
        linked resources, such as Entities, Events and Contexts.

        :param resource: The resource to register (an Intent, or a module
                         containing Intents)
        """
        if isinstance(resource, _IntentMetaclass):
            cls._register_intent(resource)

        elif isinstance(resource, ModuleType):
            for member_name, member in inspect.getmembers(resource, inspect.isclass):
                if member.__module__ == resource.__name__ and issubclass(member, Intent):
                    cls._register_intent(member)

    @classmethod
    def _register_intent(cls, intent_cls: _IntentMetaclass):
        """
        Register a single intent in the Agent class and check that language data is
        present for all supported languages (examples and responses).
        """
        if not cls.intents or not cls._intents_by_name and not cls._intents_by_event:
            assert not cls.intents and not cls._intents_by_name and not cls._intents_by_event
            assert not cls._contexts_by_name and not cls._events_by_name
            cls.intents = []
            cls._intents_by_name = {}
            cls._intents_by_event = {}
            cls._entities_by_name = {}
            cls._contexts_by_name = {}
            cls._events_by_name = {}

        if cls._intents_by_name.get(intent_cls.name):
            raise ValueError(f"Another intent exists with name {intent_cls.name}: {cls._intents_by_name[intent_cls.name]}")

        # TODO: check conflicting events
        # event_name = Agent._event_name(name)
        # if conflicting_intent := cls._intents_by_event.get(event_name):
        #     raise ValueError(f"Intent name {name} is ambiguous and clashes with {conflicting_intent} ('{conflicting_intent.metadata.name}')")

        language.intent_language_data(cls, intent_cls) # TODO: Agent languages only

        for context_cls in intent_cls.input_contexts:
            cls._register_context(context_cls)

        for context in intent_cls.output_contexts:
            cls._register_context(context)

        for event_cls in intent_cls.events:
            cls._register_event(event_cls, intent_cls)

        for param_name, param_metadata in intent_cls.parameter_schema().items():
            cls._register_entity(param_metadata.entity_cls, param_name, intent_cls.name)

        cls.intents.append(intent_cls)
        cls._intents_by_name[intent_cls.name] = intent_cls
        cls._intents_by_event[intent_cls.events[0].name] = intent_cls


    @classmethod
    def _register_entity(cls, entity_cls: _EntityMetaclass, parameter_name: str, intent_name: str):
        if not issubclass(entity_cls, EntityMixin):
            raise ValueError(f"Invalid type '{entity_cls}' for parameter '{parameter_name}' in Intent '{intent_name}': must be an Entity. Try system entities such as 'intents.Sys.Integer', or define your own custom entity.")

        if issubclass(entity_cls, SystemEntityMixin):
            return

        existing_cls = cls._entities_by_name.get(entity_cls.name)
        if not existing_cls:
            language.entity_language_data(cls, entity_cls) # Checks that language data is existing and consistent
            cls._entities_by_name[entity_cls.name] = entity_cls
            return

        if id(entity_cls) != id(existing_cls):
            existing_cls_path = f"{existing_cls.__module__}.{existing_cls.__qualname__}"
            entity_cls_path = f"{entity_cls.__module__}.{entity_cls.__qualname__}"
            raise ValueError(f"Two different Entity classes exist with the same name: '{existing_cls_path}' and '{entity_cls_path}'")

    @classmethod
    def _register_context(cls, context_obj_or_cls: Union[_ContextMetaclass, Context]):
        if isinstance(context_obj_or_cls, Context):
            context_cls = context_obj_or_cls.__class__
        elif inspect.isclass(context_obj_or_cls) and issubclass(context_obj_or_cls, Context):
            context_cls = context_obj_or_cls
        else:
            raise ValueError(f"Context {context_obj_or_cls} is not a Context instance or subclass")

        existing_cls = cls._contexts_by_name.get(context_cls.name)
        if not existing_cls:
            cls._contexts_by_name[context_cls.name] = context_cls
            return

        if id(context_cls) != id(existing_cls):
            existing_cls_path = f"{existing_cls.__module__}.{existing_cls.__qualname__}"
            context_cls_path = f"{context_cls.__module__}.{context_cls.__qualname__}"
            raise ValueError(f"Two different Context classes exist with the same name: '{existing_cls_path}' and '{context_cls_path}'")

    @classmethod
    def _register_event(cls, event_cls: _EventMetaclass, intent_cls: _IntentMetaclass):
        existing_cls = cls._events_by_name.get(event_cls.name)

        if not existing_cls:
            cls._events_by_name[event_cls.name] = event_cls
            cls._intents_by_event[event_cls.name] = intent_cls
            return

        if id(existing_cls) != id(event_cls):
            existing_cls_path = f"{existing_cls.__module__}.{existing_cls.__qualname__}"
            event_cls_path = f"{event_cls.__module__}.{event_cls.__qualname__}"
            raise ValueError(f"Two different Event classes exist with the same name: '{existing_cls_path}' and '{event_cls_path}'")

        # TODO: model different intents with same event and different input
        # context (ok) vs. different intents with same event and same input
        # context (not ok).
        existing_intent = cls._intents_by_event[event_cls.name]
        raise ValueError(f"Event '{event_cls.name}' is alreadt associated to Intent '{existing_intent}'. An Event can only be associated with 1 intent. (differenciation by input contexts is not supported yet)")

    # def save_session(self):
    #     """
    #     Store the current session (most importantly, the list of active
    #     contexts) to a persisted storage.
    #     """
    #     raise NotImplementedError("Context persistence is unsupported yet")

    # def load_session(self):
    #     """
    #     Load session information (most importantly, the list of active contexts),
    #     in a format that can be used by :meth:`Agent.predict` to restore the
    #     state before prediction.
    #     """
    #     raise NotImplementedError("Context persistence is unsupported yet")

    def _pylint_hack(self):
        raise NotImplementedError()

    @staticmethod
    def _event_name(intent_name: str) -> str:
        """
        Generate the default event name that we associate with every intent.

        >>> _event_name('test.intent_name')
        'E_TEST_INTENT_NAME'

        TODO: This is only used in Dialogflow -> Deprecate and move to DialogflowConnector
        """
        return "E_" + intent_name.upper().replace('.', '_')



import datetime

from intents import Sys
from intents.connectors.dialogflow_es.entities import TimeEntityMapping, DateEntityMapping, PersonEntityMapping

def test_time_deserialize():
    df_time = "2021-06-19T13:00:00+02:00"
    result = TimeEntityMapping().from_service(df_time)
    expected_tzinfo = datetime.timezone(datetime.timedelta(hours=2))
    assert result == Sys.Time(13, 00, tzinfo=expected_tzinfo)

def test_time_serialize():
    py_time = "11:12:00"
    result = TimeEntityMapping().to_service(py_time)
    assert result == "11:12:00"

    py_time = datetime.time(11, 12)
    result = TimeEntityMapping().to_service(py_time)
    assert result == "11:12:00"

    py_time = Sys.Time(11, 12)
    result = TimeEntityMapping().to_service(py_time)
    assert result == "11:12:00"

def test_date_deserialize():
    df_date = "2021-06-20T12:00:00+02:00"
    result = DateEntityMapping().from_service(df_date)
    assert result == Sys.Date(2021, 6, 20)

def test_date_serialize():
    py_date = "2021-07-21"
    result = DateEntityMapping().to_service(py_date)
    assert result == "2021-07-21"

    py_date = datetime.date(2021, 7, 21)
    result = DateEntityMapping().to_service(py_date)
    assert result == "2021-07-21"

    py_date = Sys.Date(2021, 7, 21)
    result = DateEntityMapping().to_service(py_date)
    assert result == "2021-07-21"

def test_person_deserialize():
    df_person = {
        "name": "John"
    }
    result = PersonEntityMapping().from_service(df_person)
    assert result == Sys.Person("John")

def test_person_serialize():
    py_person = "John"
    result = PersonEntityMapping().to_service(py_person)
    assert result == {"name": "John"}



from intents import Entity

def test_entity_default_name():

    class MyEntity(Entity):
        pass

    assert MyEntity.name == 'MyEntity'

def test_entity_custom_name():

    class MyEntity(Entity):
        name = "CustomEntityName"

    assert MyEntity.name == 'CustomEntityName'



import os
from typing import List, Dict
from dataclasses import dataclass

import yaml

from intents.model.entity import _EntityMetaclass
from intents.language.language_codes import LanguageCode
from intents.language.agent_language import agent_language_folder

@dataclass
class EntityEntry:

    value: str
    synonyms: List[str]

def entity_language_data(agent_cls: "agent._AgentMetaclass", entity_cls: _EntityMetaclass, language_code: LanguageCode=None) -> Dict[LanguageCode, List[EntityEntry]]:
    # Custom language data
    if entity_cls.custom_language_data:
        # TODO: check custom language data
        if language_code:
            return {language_code: entity_cls.custom_language_data[language_code]}
        else:
            return entity_cls.custom_language_data
    
    language_folder = agent_language_folder(agent_cls)

    if not language_code:
        result = {}
        for language_code in agent_cls.languages:
            language_data = entity_language_data(agent_cls, entity_cls, language_code)
            result[language_code] = language_data[language_code]
        return result

    language_file = os.path.join(language_folder, language_code.value, f"ENTITY_{entity_cls.name}.yaml")
    if not os.path.isfile(language_file):
        raise ValueError(f"Language file not found for entity '{entity_cls.name}'. Expected path: {language_file}.")

    with open(language_file, 'r') as f:
        language_data = yaml.load(f.read(), Loader=yaml.FullLoader)

    if not language_data:
        return []

    entries_data = language_data.get('entries', [])
    if not isinstance(entries_data, dict):
        raise ValueError(f"Invalid Entity language data for entity {entity_cls.name}. Entries data must be a dict. Entries data: {entries_data}")

    entries = []
    for value, synonyms in entries_data.items():
        if not isinstance(synonyms, list):
            raise ValueError(f"Invalid language data for entry {entity_cls.name}. Synonims data must always be lists. Synonims data for '{value}': '{synonyms}'")
        entries.append(EntityEntry(value, synonyms))

    return {language_code: entries}



"""
Here we export a :class:`Agent` to Dialogflow format.
"""
import os
import json
import shutil
import logging
import tempfile
from uuid import uuid1
from dataclasses import asdict
from typing import List, Dict, Iterable

from intents import language
from intents.model.intent import _IntentMetaclass
from intents.model.entity import SystemEntityMixin, _EntityMetaclass
import intents.connectors.dialogflow_es.agent_format as df
from intents.connectors.dialogflow_es.entities import MAPPINGS as ENTITY_MAPPINGS

logger = logging.getLogger(__name__)

def export(connector: "intents.DialogflowEsConnector", output_path: str, agent_name="py-agent") -> None:
    """
    Export the given agent to the given path
    """
    agent_cls = connector.agent_cls

    output_dir = os.path.join(tempfile.gettempdir(), 'dialogflow-agents-export', agent_name)
    intents_dir = os.path.join(output_dir, 'intents')
    entities_dir = os.path.join(output_dir, 'entities')

    if os.path.isdir(intents_dir):
        logger.warning("Removing existing intents folder: %s", intents_dir)
        shutil.rmtree(intents_dir)

    if os.path.isdir(entities_dir):
        logger.warning("Removing existing entities folder: %s", entities_dir)
        shutil.rmtree(entities_dir)

    os.makedirs(intents_dir)
    os.makedirs(entities_dir)

    with open(os.path.join(output_dir, 'agent.json'), 'w') as f:
        languages = agent_cls.languages
        json.dump(asdict(render_agent(connector, agent_name, languages)), f, indent=2)

    with open(os.path.join(output_dir, 'package.json'), 'w') as f:
        json.dump({"version": "1.0.0"}, f, indent=2)

    for intent in agent_cls.intents:
        language_data = language.intent_language_data(agent_cls, intent)
        rendered_intent = render_intent(connector, intent, language_data)
        with open(os.path.join(intents_dir, f"{intent.name}.json"), "w") as f:
            json.dump(asdict(rendered_intent), f, indent=2)
        
        for language_code, language_code_data in language_data.items():
            rendered_intent_usersays = render_intent_usersays(agent_cls, intent, language_code, language_code_data.example_utterances)
            filename = f"{intent.name}_usersays_{language_code.value}.json"
            with open(os.path.join(intents_dir, filename), "w") as f:
                usersays_data = [asdict(x) for x in rendered_intent_usersays]
                json.dump(usersays_data, f, indent=2)

    for entity_cls in agent_cls._entities_by_name.values():
        language_data = language.entity_language_data(agent_cls, entity_cls)
        rendered_entity = render_entity(entity_cls)
        with open(os.path.join(entities_dir, f"{entity_cls.name}.json"), "w") as f:
            json.dump(asdict(rendered_entity), f, indent=2)

        for language_code, entries in language_data.items():
            rendered_entity_entries = render_entity_entries(agent_cls, entries)
            filename = f"{entity_cls.name}_entries_{language_code.value}.json"
            with open(os.path.join(entities_dir, filename), "w") as f:
                entries_data = [asdict(x) for x in rendered_entity_entries]
                json.dump(entries_data, f, indent=2)

    if output_path.endswith('.zip'):
        output_path = output_path[:-4]
    shutil.make_archive(output_path, 'zip', output_dir)

#
# Agent
#

def render_agent(connector: "intents.DialogflowEsConnector",  agent_name: str, languages: List[language.LanguageCode]):
    google_assistant = df.AgentGoogleAssistant(
        project=connector.gcp_project_id,
        oAuthLinking=df.AgentGoogleAssistantOauthLinking()
        # TODO: include Google Assistant configuration
    )

    if connector.webhook_configuration:
        webhook = df.AgentWebhook(
            available=True,
            url=connector.webhook_configuration.url,
            headers=connector.webhook_configuration.headers
        )
    else:
        webhook = df.AgentWebhook()

    languages = [l.value for l in languages]
    return df.Agent(
        displayName=agent_name,
        webhook=webhook,
        googleAssistant=google_assistant,
        language=languages[0],
        supportedLanguages=languages[1:]
    )

#
# Intent
#

def render_intent(connector: "DialogflowEsConnector", intent_cls: _IntentMetaclass, language_data: Dict[language.LanguageCode, language.IntentLanguageData]):
    response = df.Response(
        affectedContexts=[df.AffectedContext(c.name, c.lifespan) for c in intent_cls.output_contexts],
        parameters=render_parameters(intent_cls, language_data),
        messages=render_responses(intent_cls, language_data, connector.rich_platforms),
    )

    return df.Intent(
        id=str(uuid1()),
        name=intent_cls.name,
        contexts=[c.name for c in intent_cls.input_contexts],
        responses=[response],

        # TODO: re-enable
        # webhookUsed=intent_cls.metadata.intent_webhook_enabled,
        # webhookForSlotFilling=intent_cls.metadata.slot_filling_webhook_enabled,
        events=[df.Event(e.name) for e in intent_cls.events]
    )

def render_parameters(intent_cls: _IntentMetaclass, language_data: Dict[language.LanguageCode, language.IntentLanguageData]):
    result = []
    for param_name, param_metadata in intent_cls.parameter_schema().items():
        entity_cls = param_metadata.entity_cls
        if issubclass(entity_cls, SystemEntityMixin):
            data_type = ENTITY_MAPPINGS[entity_cls].service_name
        else:
            data_type = entity_cls.name

        prompts = []
        for language_code, language_code_data in language_data.items():
            for prompt in language_code_data.slot_filling_prompts.get(param_name, []):
                prompts.append(df.Prompt(value=prompt, lang=language_code.value))

        result.append(df.Parameter(
            id=str(uuid1()),
            name=param_name,
            required=param_metadata.required,
            dataType=f'@{data_type}',
            value=f"${param_name}",
            defaultValue=param_metadata.default if not param_metadata.required else '',
            isList=param_metadata.is_list,
            prompts=prompts
        ))
    return result

def render_response(response: language.IntentResponse, language_code: language.LanguageCode, platform: str):
    """
    platform: None = "Default"
    """
    if isinstance(response, language.TextIntentResponse):
        response: language.TextIntentResponse
        return df.TextResponseMessage(
            lang=language_code.value,
            speech=response.choices,
            platform=platform
        )
    elif isinstance(response, language.QuickRepliesIntentResponse):
        response: language.QuickRepliesIntentResponse
        return df.QuickRepliesResponseMessage(
            lang=language_code.value,
            replies=response.replies,
            platform=platform
        )
    elif isinstance(response, language.ImageIntentResponse):
        response: language.ImageIntentResponse
        return df.ImageResponseMessage(
            lang=language_code.value,
            imageUrl=response.url,
            title=response.title if response.title else "",
            platform=platform
        )
    elif isinstance(response, language.CardIntentResponse):
        response: language.CardIntentResponse
        buttons = None
        if response.link:
            buttons = [df.CardResponseMessageButton(text="👁", postback=response.link)]
        return df.CardResponseMessage(
            lang=language_code.value,
            title=response.title,
            subtitle=response.subtitle,
            imageUrl=response.image,
            buttons=buttons,
            platform=platform
        )
    elif isinstance(response, language.CustomPayloadIntentResponse):
        response: language.CustomPayloadIntentResponse
        return df.CustomPayloadResponseMessage(
            lang=language_code.value,
            payload={
                response.name: response.payload
            },
            platform=platform
        )
    else:
        raise ValueError(f"Unsupported response type: {response}")

def render_responses(intent_cls: _IntentMetaclass, language_data: Dict[language.LanguageCode, language.IntentLanguageData], rich_platforms: Iterable[str]):
    result = []

    for language_code, language_code_data in language_data.items():
        if not language_code_data.responses:
            result.append(df.ResponseMessage(language_code.value))
            continue

        for response_group, responses in language_code_data.responses.items():
            assert response_group in [language.IntentResponseGroup.DEFAULT, language.IntentResponseGroup.RICH]
            if response_group == language.IntentResponseGroup.RICH:
                platforms_to_render = rich_platforms
            else:
                platforms_to_render = (None,) # Dialogflow will put the response in "Default" when platform=None

            for res in responses:
                for platform in platforms_to_render:
                    result.append(render_response(res, language_code, platform))

    return result

def render_utterance_chunk(chunk: language.UtteranceChunk):
    if isinstance(chunk, language.TextUtteranceChunk):
        return df.UsersaysTextChunk(text=chunk.text, userDefined=True)

    if isinstance(chunk, language.EntityUtteranceChunk):
        chunk: language.EntityUtteranceChunk
        if issubclass(chunk.entity_cls, SystemEntityMixin):
            meta = ENTITY_MAPPINGS[chunk.entity_cls].service_name
        else:
            meta = chunk.entity_cls.name

        return df.UsersaysEntityChunk(
            text=chunk.parameter_value,
            alias=chunk.parameter_name,
            meta=f'@{meta}',
            userDefined=True
        )

    raise ValueError(f"Unsupported Utterance Chunk Type: {chunk}")

def render_intent_usersays(agent_cls: type, intent: _IntentMetaclass, language_code: language.LanguageCode, examples: List[language.ExampleUtterance]):
    result = []
    for e in examples:
        result.append(df.IntentUsersays(
            id=str(uuid1()),
            lang=language_code.value,
            data=[render_utterance_chunk(c) for c in e.chunks()]
        ))
    return result

#
# Entity
#

def render_entity(entity_cls: _EntityMetaclass) -> df.Entity:
    metadata = entity_cls.metadata
    return df.Entity(
        id=str(uuid1()),
        name=entity_cls.name,
        isRegexp=metadata.regex_entity,
        automatedExpansion=metadata.automated_expansion,
        allowFuzzyExtraction=metadata.fuzzy_matching
    )

def render_entity_entries(entity_cls: _EntityMetaclass, entries: List[language.EntityEntry]) -> List[df.EntityEntry]:
    result = []
    for e in entries:
        result.append(df.EntityEntry(
            value=e.value,
            synonyms=e.synonyms
        ))
    return result




import os
import tempfile

from example_agent import ExampleAgent
from intents.connectors.dialogflow_es import export

class MockDialogflowConnector:

    agent_cls: type = ExampleAgent
    gcp_project_id: str = "fake-project-id"
    rich_platforms: tuple = ("telegram", "slack")
    webhook_configuration = None

def test_export_example_agent_no_exceptions():
    with tempfile.TemporaryDirectory() as temp_dir:
        export.export(MockDialogflowConnector(), os.path.join(temp_dir, 'TMP_AGENT.zip'))



from enum import Enum

class LanguageCode(Enum):

    ENGLISH = 'en'
    ENGLISH_US = 'en_US'
    ENGLISH_UK = 'en_UK'
    ITALIAN = 'it'
    SPANISH = 'es'
    SPANISH_SPAIN = 'es_ES'
    SPANISH_LATIN_AMERICA = 'es_LA'
    GERMAN = 'de'
    FRENCH = 'fr'
    DUTCH = 'nl'
    CHINESE = 'zh'
    CHINESE_PRC = 'zh_CN'
    CHINESE_HONG_KONG = 'zh_HK'

LANGUAGE_CODES = [x.value for x in LanguageCode]



"""
Language resources are defined separately from intent classes. They are stored
as plain YAML files in a `language/` folder, aside Agent Python modules; this is
to be flexible in the relationship with designers and translators, as well as to
allow some degree of automation when downloading cloud changes back to the local
Agent definition (this feature is currently not implemented).

Your `language/` folder will contain one subfloder per language (i.e.
`language/en/`, `language/it/`, ...); each of these will contain

* one YAML file per Intent (e.g. `my_module.my_intent_class.yaml`). Intent
  language resources are specified in :mod:`intents.language.intent_language`
* one YAML file per entity (e.g. `ENTITY_MyEntityClass.yaml`). Entity language
  resources are specified in :mod:`intents.language.entity_language`

.. tip::

    It may be useful to look at The Example Agent code and language resources (https://github.com/dariowho/intents/tree/master/example_agent/language) to get more insight on the format and naming conventions of language files.
"""
from intents.language.language_codes import LanguageCode, LANGUAGE_CODES
from intents.language.agent_language import agent_language_folder, agent_supported_languages
from intents.language.intent_language import intent_language_data, IntentResponseGroup, IntentResponse, TextIntentResponse, ImageIntentResponse, QuickRepliesIntentResponse, CardIntentResponse, CustomPayloadIntentResponse, IntentLanguageData, ExampleUtterance, UtteranceChunk, TextUtteranceChunk, EntityUtteranceChunk
from intents.language.entity_language import entity_language_data, EntityEntry



import logging
import datetime

from intents.model.entity import Sys
from intents.service_connector import EntityMapping, StringEntityMapping, ServiceEntityMappings

class PersonEntityMapping(EntityMapping):
    """
    Dialogflow's `sys.person` entity type is returned as an object like
    `{"name": "Guido"}`
    """

    entity_cls = Sys.Person
    service_name = 'sys.person'

    def from_service(self, service_data: dict) -> Sys.Person:
        if isinstance(service_data, str):
            service_data = {"name": service_data}
        return Sys.Person(service_data['name'])

    def to_service(self, entity: Sys.Person) -> dict:
        return {"name": str(entity)}

class TimeEntityMapping(EntityMapping):
    """
    Dialogflow's `sys.time` entity is returned as a full ISO datetime string
    (e.g. "2021-06-19T13:00:00+02:00"). We only consider the time part, which is
    returned as a :class:`datetime.time` object. The timezone is the same as the
    one that is returned by Dialogflow.

    Both :class:`datetime.time` and :class:`str` can be serialized to be used as
    parameter values. As :class:`Sys.Time` inherits from :class:`datetime.time`,
    these are all valid usages:

    >>> connector.trigger(lunch_time(time=Sys.Time(12, 30)))
    >>> connector.trigger(lunch_time(time=datetime.time(12, 30)))
    >>> connector.trigger(lunch_time(time="12:30:00"))

    Note that strings will be sent to Dialogflow without checks: be sure to use
    valid time references. Also, note that Dialogflow won't resolve relative
    time references in triggers. The following code will result in an error:

    >>> connector.trigger(lunch_time(time="in 10 minutes"))
    RuntimeError: ...
    """

    entity_cls = Sys.Time
    service_name = 'sys.time'

    def from_service(self, service_data: str) -> Sys.Time:
        dt = datetime.datetime.fromisoformat(service_data)
        return Sys.Time.from_py_time(dt.timetz())

    def to_service(self, entity: Sys.Time) -> str:
        # TODO: check
        return str(entity)

class DateEntityMapping(EntityMapping):
    """
    Dialogflow's `sys.date` entity is returned as a full ISO datetime string
    (e.g. "2021-06-19T13:00:00+02:00"). We only consider the date part, which is
    returned as a :class:`datetime.date` object (specifically, :class:`Sys.Date`).

    Both :class:`datetime.date` and :class:`str` can be serialized to be used as
    parameter values. As :class:`Sys.Date` inherits from :class:`datetime.date`,
    these are all valid usages:

    >>> connector.trigger(birthday_intent(date=Sys.Date(2021, 07, 21)))
    >>> connector.trigger(birthday_intent(date=datetime.date(2021, 07, 21)))
    >>> connector.trigger(birthday_intent(date="2021-07-21"))

    Note that strings will be sent to Dialogflow without checks: be sure to use
    valid date references. Also, note that Dialogflow won't resolve relative
    date references in triggers. The following code will result in an error:

    >>> connector.trigger(birthday_intent(date="tomorrow"))
    RuntimeError: ...
    """

    entity_cls = Sys.Date
    service_name = 'sys.date'

    def from_service(self, service_data: str) -> Sys.Date:
        df_datetime = datetime.datetime.fromisoformat(service_data)
        return Sys.Date.from_py_date(df_datetime.date())

    def to_service(self, entity: datetime.date) -> str:
        return str(entity)


MAPPINGS = ServiceEntityMappings.from_list([
    # StringEntityMapping(Sys.Any, "sys.any"),
    DateEntityMapping(),
    TimeEntityMapping(),
    StringEntityMapping(Sys.Integer, "sys.number-integer"),
    StringEntityMapping(Sys.Email, "sys.email"),
    StringEntityMapping(Sys.PhoneNumber, "sys.phone-number"),
    StringEntityMapping(Sys.Color, "sys.color"),
    StringEntityMapping(Sys.Language, "sys.language"),
    StringEntityMapping(Sys.Url, "sys.url"),
    PersonEntityMapping(),
    StringEntityMapping(Sys.MusicArtist, "sys.music-artist"),
    StringEntityMapping(Sys.MusicGenre, "sys.music-genre"),
])



from unittest.mock import patch, call

import pytest

from intents import Agent, Sys, Intent, Entity
from intents import language
from example_agent import smalltalk

real_LanguageCode = language.LanguageCode

def _get_toy_agent():

    class ToyAgent(Agent):
        languages = ['en']

    return ToyAgent

@patch('intents.language.intent_language_data')
def test_register_intent_valid_language_data(mock_language):
    mock_language.LanguageCode = real_LanguageCode

    MyAgent = _get_toy_agent()
    MyAgent._register_intent(smalltalk.hello)

    assert MyAgent.intents == [smalltalk.hello]
    assert MyAgent._intents_by_name == {'smalltalk.hello': smalltalk.hello}
    assert MyAgent._intents_by_event == {'E_SMALLTALK_HELLO': smalltalk.hello}

    MyAgent = _get_toy_agent()
    MyAgent.register(smalltalk.hello)

    assert MyAgent.intents == [smalltalk.hello]
    assert MyAgent._intents_by_name == {'smalltalk.hello': smalltalk.hello}
    assert MyAgent._intents_by_event == {'E_SMALLTALK_HELLO': smalltalk.hello}

@patch('intents.language.intent_language_data')
def test_register_intent_invalid_language_data(mock_language):
    mock_language.LanguageCode = real_LanguageCode

    mock_language.side_effect = ValueError("Mock language data not found exception")

    MyAgent = _get_toy_agent()
    
    with pytest.raises(ValueError):
        MyAgent._register_intent(smalltalk.hello)

@patch('intents.model.agent.language')
def test_register_intent_registers_entities(mock_language):
    mock_language.LanguageCode = real_LanguageCode

    MyAgent = _get_toy_agent()

    class MyEntity(Entity):
        pass

    class MyIntent(Intent):
        name = "my_intent"

        a_param: Sys.Person
        another_param: MyEntity

    MyAgent._register_intent(MyIntent)

    assert MyAgent.intents == [MyIntent]
    assert MyAgent._intents_by_name == {'my_intent': MyIntent}
    assert MyAgent._intents_by_event == {'E_MY_INTENT': MyIntent}
    assert MyAgent._entities_by_name == {'MyEntity': MyEntity}

    class MyOtherIntent(Intent):
        name = "my_other_intent"

        a_param: Sys.Person
        another_param: MyEntity

    MyAgent._register_intent(MyOtherIntent)
    assert MyAgent.intents == [MyIntent, MyOtherIntent]
    assert MyAgent._intents_by_name == {'my_intent': MyIntent, 'my_other_intent': MyOtherIntent}
    assert MyAgent._intents_by_event == {'E_MY_INTENT': MyIntent, 'E_MY_OTHER_INTENT': MyOtherIntent}
    assert MyAgent._entities_by_name == {'MyEntity': MyEntity} # MyEntity is not registered twice
    
    def _make_duplicate_entity():
        class MyEntity(Entity):
            """Same name, different class"""
            pass
        return MyEntity

    class MyInvalidIntent(Intent):
        name = "my_invalid_intent"

        a_param: Sys.Person
        another_param: _make_duplicate_entity()

    with pytest.raises(ValueError):
        MyAgent._register_intent(MyInvalidIntent) # Different entities with same name raise Error

@patch('intents.model.agent.language')
def test_register_module(mock_language):
    mock_language.LanguageCode = real_LanguageCode
    MyAgent = _get_toy_agent()
    with patch.object(MyAgent, '_register_intent') as mock_register_intent:
        MyAgent.register(smalltalk)
        mock_register_intent.asssert_has_calls([
            call(MyAgent, smalltalk.hello),
            call(MyAgent, smalltalk.user_name_give),
            call(MyAgent, smalltalk.agent_name_give),
            call(MyAgent, smalltalk.user_likes_music),
            call(MyAgent, smalltalk.greet_friends),
            call(MyAgent, smalltalk.agent_welcomes_user),
        ])

# test duplicate intent
